@@ -22,9 +22,13 @@ pub trait Value: Clone + From<DataValue> {
     fn bool(b: bool, ty: Type) -> ValueResult<Self>;
     fn into_bool(self) -> ValueResult<bool>;
     fn vector(v: [u8; 16], ty: Type) -> ValueResult<Self>;
+    fn into_array(&self) -> ValueResult<[u8; 16]>;
     fn convert(self, kind: ValueConversionKind) -> ValueResult<Self>;
     fn concat(self, other: Self) -> ValueResult<Self>;
 
+    fn max(self, other: Self) -> ValueResult<Self>;
+    fn min(self, other: Self) -> ValueResult<Self>;
+
     // Comparison.
     fn eq(&self, other: &Self) -> ValueResult<bool>;
     fn gt(&self, other: &Self) -> ValueResult<bool>;
@@ -47,6 +51,10 @@ pub trait Value: Clone + From<DataValue> {
     fn div(self, other: Self) -> ValueResult<Self>;
     fn rem(self, other: Self) -> ValueResult<Self>;
 
+    // Saturating arithmetic.
+    fn add_sat(self, other: Self) -> ValueResult<Self>;
+    fn sub_sat(self, other: Self) -> ValueResult<Self>;
+
     // Bitwise.
     fn shl(self, other: Self) -> ValueResult<Self>;
     fn ushr(self, other: Self) -> ValueResult<Self>;
@@ -57,6 +65,13 @@ pub trait Value: Clone + From<DataValue> {
     fn or(self, other: Self) -> ValueResult<Self>;
     fn xor(self, other: Self) -> ValueResult<Self>;
     fn not(self) -> ValueResult<Self>;
+
+    // Bit counting.
+    fn count_ones(self) -> ValueResult<Self>;
+    fn leading_ones(self) -> ValueResult<Self>;
+    fn leading_zeros(self) -> ValueResult<Self>;
+    fn trailing_zeros(self) -> ValueResult<Self>;
+    fn reverse_bits(self) -> ValueResult<Self>;
 }
 
 #[derive(Error, Debug, PartialEq)]
@@ -71,6 +86,8 @@ pub enum ValueError {
     InvalidDataValueCast(#[from] DataValueCastFailure),
     #[error("performed a division by zero")]
     IntegerDivisionByZero,
+    #[error("performed a operation that overflowed this integer type")]
+    IntegerOverflow,
 }
 
 #[derive(Debug, PartialEq)]
@@ -78,6 +95,7 @@ pub enum ValueTypeClass {
     Integer,
     Boolean,
     Float,
+    Vector,
 }
 
 impl Display for ValueTypeClass {
@@ -86,11 +104,12 @@ impl Display for ValueTypeClass {
             ValueTypeClass::Integer => write!(f, "integer"),
             ValueTypeClass::Boolean => write!(f, "boolean"),
             ValueTypeClass::Float => write!(f, "float"),
+            ValueTypeClass::Vector => write!(f, "vector"),
         }
     }
 }
 
-#[derive(Debug)]
+#[derive(Debug, Clone)]
 pub enum ValueConversionKind {
     /// Throw a [ValueError] if an exact conversion to [Type] is not possible; e.g. in `i32` to
     /// `i16`, convert `0x00001234` to `0x1234`.
@@ -123,6 +142,20 @@ pub enum ValueConversionKind {
 
 /// Helper for creating match expressions over [DataValue].
 macro_rules! unary_match {
+    ( $op:ident($arg1:expr); [ $( $data_value_ty:ident ),* ]; [ $( $return_value_ty:ident ),* ] ) => {
+        match $arg1 {
+            $( DataValue::$data_value_ty(a) => {
+                Ok(DataValue::$data_value_ty($return_value_ty::try_from(a.$op()).unwrap()))
+            } )*
+            _ => unimplemented!()
+        }
+    };
+    ( $op:ident($arg1:expr); [ $( $data_value_ty:ident ),* ] ) => {
+        match $arg1 {
+            $( DataValue::$data_value_ty(a) => { Ok(DataValue::$data_value_ty(a.$op())) } )*
+            _ => unimplemented!()
+        }
+    };
     ( $op:tt($arg1:expr); [ $( $data_value_ty:ident ),* ] ) => {
         match $arg1 {
             $( DataValue::$data_value_ty(a) => { Ok(DataValue::$data_value_ty($op a)) } )*
@@ -223,8 +256,16 @@ impl Value for DataValue {
         }
     }
 
-    fn vector(_v: [u8; 16], _ty: Type) -> ValueResult<Self> {
-        unimplemented!()
+    fn vector(v: [u8; 16], ty: Type) -> ValueResult<Self> {
+        assert!(ty.is_vector() && ty.bytes() == 16);
+        Ok(DataValue::V128(v))
+    }
+
+    fn into_array(&self) -> ValueResult<[u8; 16]> {
+        match *self {
+            DataValue::V128(v) => Ok(v),
+            _ => Err(ValueError::InvalidType(ValueTypeClass::Vector, self.ty())),
+        }
     }
 
     fn convert(self, kind: ValueConversionKind) -> ValueResult<Self> {
@@ -232,7 +273,19 @@ impl Value for DataValue {
             ValueConversionKind::Exact(ty) => match (self, ty) {
                 // TODO a lot to do here: from bmask to ireduce to raw_bitcast...
                 (DataValue::I64(n), types::I32) => DataValue::I32(i32::try_from(n)?),
+                (DataValue::I64(n), types::I64) => DataValue::I64(n),
+                (DataValue::I64(n), types::I128) => DataValue::I128(n as i128),
                 (DataValue::B(b), t) if t.is_bool() => DataValue::B(b),
+                (DataValue::B(b), t) if t.is_int() => {
+                    // Bools are represented in memory as all 1's
+                    let val = match (b, t) {
+                        (true, types::I128) => -1,
+                        (true, t) => (1i128 << t.bits()) - 1,
+                        _ => 0,
+                    };
+                    DataValue::int(val, t)?
+                }
+                (dv, t) if t.is_int() && dv.ty() == t => dv,
                 (dv, _) => unimplemented!("conversion: {} -> {:?}", dv.ty(), kind),
             },
             ValueConversionKind::Truncate(ty) => {
@@ -255,36 +308,52 @@ impl Value for DataValue {
                     kind
                 );
 
-                let shift_amt = 128 - (ty.bytes() * 8);
+                let shift_amt = (self.ty().bytes() * 8) - (ty.bytes() * 8);
                 let mask = (1 << (ty.bytes() * 8)) - 1i128;
                 let shifted_mask = mask << shift_amt;
 
                 let extracted = (self.into_int()? & shifted_mask) >> shift_amt;
                 Self::from_integer(extracted, ty)?
             }
-            ValueConversionKind::SignExtend(ty) => match (self.ty(), ty) {
-                (types::I8, types::I16) => unimplemented!(),
-                (types::I8, types::I32) => unimplemented!(),
-                (types::I8, types::I64) => unimplemented!(),
-                (types::I16, types::I32) => unimplemented!(),
-                (types::I16, types::I64) => unimplemented!(),
-                (types::I32, types::I64) => unimplemented!(),
-                _ => unimplemented!("conversion: {} -> {:?}", self.ty(), kind),
+            ValueConversionKind::SignExtend(ty) => match (self, ty) {
+                (DataValue::U8(n), types::I16) => DataValue::U16(n as u16),
+                (DataValue::U8(n), types::I32) => DataValue::U32(n as u32),
+                (DataValue::U8(n), types::I64) => DataValue::U64(n as u64),
+                (DataValue::I8(n), types::I16) => DataValue::I16(n as i16),
+                (DataValue::I8(n), types::I32) => DataValue::I32(n as i32),
+                (DataValue::I8(n), types::I64) => DataValue::I64(n as i64),
+                (DataValue::U16(n), types::I32) => DataValue::U32(n as u32),
+                (DataValue::U16(n), types::I64) => DataValue::U64(n as u64),
+                (DataValue::I16(n), types::I32) => DataValue::I32(n as i32),
+                (DataValue::I16(n), types::I64) => DataValue::I64(n as i64),
+                (DataValue::U32(n), types::I64) => DataValue::U64(n as u64),
+                (DataValue::I32(n), types::I64) => DataValue::I64(n as i64),
+                (DataValue::I64(n), types::I128) => DataValue::I128(n as i128),
+                (dv, _) => unimplemented!("conversion: {} -> {:?}", dv.ty(), kind),
             },
-            ValueConversionKind::ZeroExtend(ty) => match (self.ty(), ty) {
-                (types::I8, types::I16) => unimplemented!(),
-                (types::I8, types::I32) => unimplemented!(),
-                (types::I8, types::I64) => unimplemented!(),
-                (types::I16, types::I32) => unimplemented!(),
-                (types::I16, types::I64) => unimplemented!(),
-                (types::I32, types::I64) => unimplemented!(),
-                _ => unimplemented!("conversion: {} -> {:?}", self.ty(), kind),
+            ValueConversionKind::ZeroExtend(ty) => match (self, ty) {
+                (DataValue::U8(n), types::I16) => DataValue::U16(n as u16),
+                (DataValue::U8(n), types::I32) => DataValue::U32(n as u32),
+                (DataValue::U8(n), types::I64) => DataValue::U64(n as u64),
+                (DataValue::I8(n), types::I16) => DataValue::I16(n as u8 as i16),
+                (DataValue::I8(n), types::I32) => DataValue::I32(n as u8 as i32),
+                (DataValue::I8(n), types::I64) => DataValue::I64(n as u8 as i64),
+                (DataValue::U16(n), types::I32) => DataValue::U32(n as u32),
+                (DataValue::U16(n), types::I64) => DataValue::U64(n as u64),
+                (DataValue::I16(n), types::I32) => DataValue::I32(n as u16 as i32),
+                (DataValue::I16(n), types::I64) => DataValue::I64(n as u16 as i64),
+                (DataValue::U32(n), types::I64) => DataValue::U64(n as u64),
+                (DataValue::I32(n), types::I64) => DataValue::I64(n as u32 as i64),
+                (DataValue::I64(n), types::I128) => DataValue::I128(n as u64 as i128),
+                (from, to) if from.ty() == to => from,
+                (dv, _) => unimplemented!("conversion: {} -> {:?}", dv.ty(), kind),
             },
             ValueConversionKind::ToUnsigned => match self {
                 DataValue::I8(n) => DataValue::U8(n as u8),
                 DataValue::I16(n) => DataValue::U16(n as u16),
                 DataValue::I32(n) => DataValue::U32(n as u32),
                 DataValue::I64(n) => DataValue::U64(n as u64),
+                DataValue::I128(n) => DataValue::U128(n as u128),
                 _ => unimplemented!("conversion: {} -> {:?}", self.ty(), kind),
             },
             ValueConversionKind::ToSigned => match self {
@@ -292,6 +361,7 @@ impl Value for DataValue {
                 DataValue::U16(n) => DataValue::I16(n as i16),
                 DataValue::U32(n) => DataValue::I32(n as i32),
                 DataValue::U64(n) => DataValue::I64(n as i64),
+                DataValue::U128(n) => DataValue::I128(n as i128),
                 _ => unimplemented!("conversion: {} -> {:?}", self.ty(), kind),
             },
             ValueConversionKind::RoundNearestEven(ty) => match (self.ty(), ty) {
@@ -315,12 +385,28 @@ impl Value for DataValue {
         }
     }
 
+    fn max(self, other: Self) -> ValueResult<Self> {
+        if Value::gt(&self, &other)? {
+            Ok(self)
+        } else {
+            Ok(other)
+        }
+    }
+
+    fn min(self, other: Self) -> ValueResult<Self> {
+        if Value::lt(&self, &other)? {
+            Ok(self)
+        } else {
+            Ok(other)
+        }
+    }
+
     fn eq(&self, other: &Self) -> ValueResult<bool> {
-        comparison_match!(PartialEq::eq[&self, &other]; [I8, I16, I32, I64, U8, U16, U32, U64, F32, F64])
+        comparison_match!(PartialEq::eq[&self, &other]; [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128, F32, F64])
     }
 
     fn gt(&self, other: &Self) -> ValueResult<bool> {
-        comparison_match!(PartialOrd::gt[&self, &other]; [I8, I16, I32, I64, U8, U16, U32, U64, F32, F64])
+        comparison_match!(PartialOrd::gt[&self, &other]; [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128, F32, F64])
     }
 
     fn uno(&self, other: &Self) -> ValueResult<bool> {
@@ -333,12 +419,14 @@ impl Value for DataValue {
             (DataValue::I16(a), DataValue::I16(b)) => a.checked_sub(*b).is_none(),
             (DataValue::I32(a), DataValue::I32(b)) => a.checked_sub(*b).is_none(),
             (DataValue::I64(a), DataValue::I64(b)) => a.checked_sub(*b).is_none(),
+            (DataValue::I128(a), DataValue::I128(b)) => a.checked_sub(*b).is_none(),
             _ => unimplemented!(),
         })
     }
 
     fn add(self, other: Self) -> ValueResult<Self> {
-        binary_match!(wrapping_add(&self, &other); [I8, I16, I32, I64, I128]) // TODO: floats must handle NaNs, +/-0
+        // TODO: floats must handle NaNs, +/-0
+        binary_match!(wrapping_add(&self, &other); [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128])
     }
 
     fn sub(self, other: Self) -> ValueResult<Self> {
@@ -350,7 +438,15 @@ impl Value for DataValue {
     }
 
     fn div(self, other: Self) -> ValueResult<Self> {
-        if other.clone().into_int()? == 0 {
+        let denominator = other.clone().into_int()?;
+
+        // Check if we are dividing INT_MIN / -1. This causes an integer overflow trap.
+        let min = Value::int(1i128 << (self.ty().bits() - 1), self.ty())?;
+        if self == min && denominator == -1 {
+            return Err(ValueError::IntegerOverflow);
+        }
+
+        if denominator == 0 {
             return Err(ValueError::IntegerDivisionByZero);
         }
 
@@ -365,6 +461,14 @@ impl Value for DataValue {
         binary_match!(%(&self, &other); [I8, I16, I32, I64])
     }
 
+    fn add_sat(self, other: Self) -> ValueResult<Self> {
+        binary_match!(saturating_add(self, &other); [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128])
+    }
+
+    fn sub_sat(self, other: Self) -> ValueResult<Self> {
+        binary_match!(saturating_sub(self, &other); [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128])
+    }
+
     fn shl(self, other: Self) -> ValueResult<Self> {
         binary_match!(<<(&self, &other); [I8, I16, I32, I64])
     }
@@ -386,11 +490,11 @@ impl Value for DataValue {
     }
 
     fn and(self, other: Self) -> ValueResult<Self> {
-        binary_match!(&(&self, &other); [I8, I16, I32, I64])
+        binary_match!(&(&self, &other); [B, I8, I16, I32, I64])
     }
 
     fn or(self, other: Self) -> ValueResult<Self> {
-        binary_match!(|(&self, &other); [I8, I16, I32, I64])
+        binary_match!(|(&self, &other); [B, I8, I16, I32, I64])
     }
 
     fn xor(self, other: Self) -> ValueResult<Self> {
@@ -400,4 +504,24 @@ impl Value for DataValue {
     fn not(self) -> ValueResult<Self> {
         unary_match!(!(&self); [I8, I16, I32, I64])
     }
+
+    fn count_ones(self) -> ValueResult<Self> {
+        unary_match!(count_ones(&self); [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128]; [i8, i16, i32, i64, i128, u8, u16, u32, u64, u128])
+    }
+
+    fn leading_ones(self) -> ValueResult<Self> {
+        unary_match!(leading_ones(&self); [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128]; [i8, i16, i32, i64, i128, u8, u16, u32, u64, u128])
+    }
+
+    fn leading_zeros(self) -> ValueResult<Self> {
+        unary_match!(leading_zeros(&self); [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128]; [i8, i16, i32, i64, i128, u8, u16, u32, u64, u128])
+    }
+
+    fn trailing_zeros(self) -> ValueResult<Self> {
+        unary_match!(trailing_zeros(&self); [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128]; [i8, i16, i32, i64, i128, u8, u16, u32, u64, u128])
+    }
+
+    fn reverse_bits(self) -> ValueResult<Self> {
+        unary_match!(reverse_bits(&self); [I8, I16, I32, I64, I128, U8, U16, U32, U64, U128])
+    }
 }
